# -*- coding: utf-8 -*-
# Copyright (C) 2019 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.modules.backups"""

import logging
import os
import threading
import time

from datetime import datetime

import six

from dwho.adapters.redis import DWhoAdapterRedis
from dwho.classes.errors import DWhoConfigurationError
from dwho.classes.modules import DWhoModuleBase, MODULES
from dwho.classes.inotify import DWhoInotifyConfig
from dwho.classes.notifiers import DWhoPushNotifications
from httpdis.ext.httpdis_json import HttpReqErrJson
from sonicprobe.libs import anysql, mysql_config_parser, urisup
from sonicprobe.libs.moresynchro import RWLock

from ..classes.common import (DATE_FORMAT,
                              STATUS_FAILURE,
                              STATUS_SUCCESS,
                              STATUS_STOPPED,
                              STATUS_UNKNOWN)
from ..classes.lock_db import BackupsLockDB
from ..classes.scan_paths import BackupsScanPaths
from ..classes.s3 import XtrargonS3Helper


LOG                         = logging.getLogger('xtrargon_ng.modules.backups')

DB_URI_PARAMS               = (('autocommit', '1'),
                               ('charset', 'utf8'),
                               ('cursorclass', 'Dict'),
                               ('wait_timeout', '2147483'))


class BackupsModule(DWhoModuleBase):
    MODULE_NAME = 'backups'

    LOCK = RWLock()

    @staticmethod
    def _load_path_params(params):
        r = {'follow': True,
             'exclude_patterns': None,
             'dest_params': None}

        if not params or not isinstance(params, dict):
            return r

        r['follow'] = bool(params.get('follow', True))

        if 'exclude_files' in params:
            r['exclude_patterns'] = DWhoInotifyConfig.load_exclude_patterns(params['exclude_files'])

        if isinstance(params.get('dest_params'), dict):
            r['dest_params'] = params['dest_params'].copy()

        return r

    def _load_s3helper(self):
        if not self.config.get('backups'):
            raise DWhoConfigurationError("missing backups configuration")

        for x in ('paths', 's3'):
            if x not in self.config['backups'] \
             or not isinstance(self.config['backups'][x], dict):
                raise DWhoConfigurationError("missing backups %s" % x)

        for toppath, params in six.iteritems(self.config['backups']['paths'].copy()):
            self.config['backups']['paths'][toppath] = self._load_path_params(params)

        r = XtrargonS3Helper(self.config['backups']['s3'])
        r.load_credentials()

        if not r.get_bucket():
            raise DWhoConfigurationError("missing default bucket setting")

        return r

    def _build_mysql_uri(self):
        host = mysql_config_parser.MYSQL_DEFAULT_HOST
        port = mysql_config_parser.MYSQL_DEFAULT_PORT

        if self.dbconf.has_option('client', 'host'):
            host = self.dbconf.get('client', 'host')

        if self.dbconf.has_option('client', 'port'):
            port = self.dbconf.get('client', 'port')

        uri = ('mysql',
               (self.dbconf.get('client', 'user'),
                self.dbconf.get('client', 'password'),
                host,
                port),
               None,
               DB_URI_PARAMS,
               None)

        return urisup.uri_help_unsplit(uri)

    @staticmethod
    def _check_mysql_server(uri):
        dbconn = None
        try:
            dbconn = anysql.connect_by_uri(uri)
            if not dbconn.is_connected():
                raise DWhoConfigurationError("unable to connect to mysql server")
        finally:
            if dbconn:
                dbconn.close()

    def _check_redis_server(self, prefix):
        redisconn = DWhoAdapterRedis(self.config, prefix = prefix)
        if not redisconn.servers:
            raise DWhoConfigurationError("missing redis server for %s" % prefix)

        redisconn.ping()

        return redisconn

    # pylint: disable-msg=attribute-defined-outside-init
    def safe_init(self, options):
        self.lock_timeout = self.config['general']['lock_timeout']

        self.is_running   = threading.Event()
        self.is_scanning  = threading.Event()
        self.lock_db      = threading.Event()
        self.db_locked    = threading.Event()

        self._bkp_scan    = None
        self._db_lock     = None
        self._errors      = []
        self._state       = {}

        self.dbconf       = mysql_config_parser.MySQLConfigVersion(
            custom_file = os.path.join('client.d', "%s.cnf" % self.server_id),
            config_dir  = self.config.get('_config_directory')).get_client()
        self.dburi        = self._build_mysql_uri()

        self._check_mysql_server(self.dburi)
        self._bkp_history = self._check_redis_server('backups_history')
        self.file_sync    = self._check_redis_server('file_sync')

        self.s3helper     = self._load_s3helper()

        self.notifiers    = None
        if self.config['general'].get('notifiers_path'):
            self.notifiers = DWhoPushNotifications(self.server_id,
                                                   self.config['general']['notifiers_path'])

    def _stop(self):
        if self._bkp_scan:
            self._bkp_scan.terminate()
            self._bkp_scan = None

        if self._db_lock:
            self._db_lock.terminate()
            self._db_lock = None

    def at_stop(self):
        self._stop()

    def _start(self, start = True, clear_events = True):
        self._stop()
        self.reset_state()

        self._db_lock  = BackupsLockDB(self)
        self._db_lock.daemon = True

        self._bkp_scan = BackupsScanPaths(self)
        self._bkp_scan.daemon = True

        if clear_events:
            self.is_running.clear()
            self.is_scanning.clear()
            self.lock_db.clear()
            self.db_locked.clear()

        if start:
            self._bkp_scan.start()
            self._db_lock.start()
            self.is_running.set()

    def add_errors(self, errors):
        if isinstance(errors, six.string_types):
            errors = [errors]

        if isinstance(errors, list):
            self._errors.extend(errors)

    def has_errors(self):
        return len(self._errors) > 0

    def get_state(self, status = None, end = False):
        self._state['errors'] = self._errors

        if status:
            self._state['status'] = status

        if not end or self._state['ended_at']:
            return self._state

        started_at = self._state.get('started_at')
        if started_at:
            started_at = datetime.strptime(started_at, DATE_FORMAT)
            duration   = (datetime.now() - started_at).total_seconds()
            self._state['duration'] = duration

        self._state['ended_at'] = time.strftime(DATE_FORMAT)

        if self._state['status'] == STATUS_STOPPED:
            self._state['errors'].append('stopped before completing')
        elif not self.has_errors():
            self._state['status'] = STATUS_SUCCESS
        else:
            self._state['status'] = STATUS_FAILURE

        return self._state

    def reset_state(self):
        self._state = {'status':     STATUS_UNKNOWN,
                       'started_at': None,
                       'ended_at':   None,
                       'uuid':       None,
                       'duration':   -1,
                       'errors':     []}

    def get_state_value(self, key):
        return self._state.get(key)

    def set_state_value(self, key, value):
        self._state[key] = value

    def start(self, request): # pylint: disable-msg=unused-argument
        if self.is_running.is_set():
            raise HttpReqErrJson(409, "backup is already processing")

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            self._start()
            self.is_scanning.wait(self.lock_timeout)
            return self.get_state()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def stop(self, request): # pylint: disable-msg=unused-argument
        if not self.is_running.is_set():
            raise HttpReqErrJson(409, "backup isn't running")

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            self._stop()
            xtime = time.time()

            while self.is_running.is_set():
                if time.time() - xtime > self.lock_timeout:
                    self.add_errors("unable to stop backup")
                    break
                time.sleep(0.1)
            return self.get_state(end = True)
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(BackupsModule())
    _start()
